//
//						CPU16.v - 16-bit Central Processing Unit
//
//					  (C) Copyright 2007-2008 John B. Stephensen
//
// This Verilog source file and all its derivatives are licensed only for
// personal non-profit educational use in the Amateur Radio Service and
// the license is not transferrable. The information is provided as-is for
// experimental purposes and the author does not warranty its freedom
// from defects or its suitability for any specific application.
//
// This CPU is a RISC processor with 16 16-bit registers using 2-address instructions. It
// uses a Harvard architecture with a 16-bit instruction size and 16-bit data word lengths.
// A subsidiary module decodes program flow control instructions (type 0). Data manipulation
// instruction formats are:
// 
//		 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 0 |                                                           | Program flow
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 0   1 | 1   1 |             Data              |   Register A  | MVI
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 0   0 |             Data              |   Register A  | ADI
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 0   1 |             Data              |   Register A  | CPI
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 1   0   0 |    MOD    |   Register B  |   Register A  | MUL
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 1   0   1 | 0   0 |MOD|   Register B  |   Register A  | DIV
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 1   0   1 | 1   0   x |   Bit Select  |   Register A  | TST
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 1   0   1 | 1   1   x | x   x   x   C | x   x   x   x | STC/CLC
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   0 | 1   1   0 |    MOD    |   Register B  |   Register A  | COMP
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
// 	| 1   0 | 1   1   1 |        Port Number        | Data Register | OUT
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 0   0   0 |        Port Number        | Data Register | IN
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 0   0   1 | x | R | W |    Address    | Data Register | R/W
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 0   1   0 |    MOD    | Reg. B/Const. |   Register A  | 16-bit Math
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 0   1   1 | 0 |  MOD  | Reg. B/Const. |   Register A  | 8-bit Math
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 0   1   1 | 1 |  MOD  | Reg. B/Const. |   Register A  | 4-bit Math
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 1   0   0 |    MOD    | Reg. B/Const. |   Register A  | Logic
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 1   0   1 |    MOD    |   Register B  |   Register A  | SXL/SXH/RR4/RR8/RR12/REV
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 1   0   0 | 1   1   0 |      MOD      |   Register A  |	SHR
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 1   0   0 | 1   1   1 |      MOD      |   Register A  |	SHL
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
//		| 1   1 | 1   1   0 |   MOD     | x   x   x   x |   Register A  | LOAD P/Q/R
//		+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
// 
//	MATH	Scalar arithmetic operations requiring one clock cycle.
//	LOGIC	Logic operations requiring one clock cycle.
//	CMATH	Complex arithmetic operations requiring one clock cycle.
// R/W	Swap contents of register A with memory location addressed by register B or just read or write.
//	I/O	Access I/O port specified by the B field and write register A contents or read into register A.
// LOAD	Load multiply and divide results into registers (LPL, LPH, LQ, LR)
// LDL	Load register A with 8-bit constant and extend sign.
// LDH	Load upper half of register A with 8-bit constant and preserve lower 8 bits.
//	TEST	Test individual bits within register A.
//	COMP	Compare two registers or compare one register to zero.
//	OP		Operation type uses multiplexer to select functional unit.
//	MOD	Modifier selects operation type within functional unit.
//
// Bits 15 and 14 encode the instruction type. A type 1 instruction carries immediate data in bits 11-4
// and specifies the register in bits 3-0. In type 2 and 3 instructions, bits 13-11 select a functional
// unit in the ALU to carry out a particular type of operation. Bits 10-8 specify the details of the
// operation performed by that functional unit. Bits 7-4 specify a source register or contain a constant
// value. Bits 3-0 specify the destination register which may also be a source. I/O instructions are an
// exception where bits 10-4 specify the I/O port being accessed. Memory instructions use the contents
// of register B as the address. Both use register A as the data source for writes and the data
// destination for reads. During memory access, the R bit enables reading and the W bit enables writing.
// This module also determines the flags tested for conditional jumps as follows:
//
//		Flag	Condition
//		----	------------------------------
//		 00	Unconditional
//		 01	Overflow flag
//		 10	Carry flag
//		 11	Zero flag
//
// The carry and zero flags are only modified by 16-bit add, subtract, increment and decrement operations.
// Only bit test and shift instructions set or reset the link flag. All instructions execute in one clock
// cycle.
//
// Without multipler and divider: 101 MHz maximum and uses 240 slices (4% XC3S500E).
// With multipler and divider: 101 MHz maximum and uses 1 multiplier and 327 slices (6% XC3S500E).
// Speed limited by B input to adder/subtractor. Next slowest is MAC and condition code checking.
//
//	Normal warnings:
//		Signals <prod<35:34>> and <prod<1:0>> are assigned but never used.
//
// History:
//		M001	5-31-08	Load immediate data to register.
//		M002	6-8-08	Remove encode instruction
//		M003	8-5-08	Convert to 16-bit instruction format
//		M004	9-29-08	Move input multiplexer to MCU16.v and bit test to ALU.v
//		M005	2-28-09	Use external memory with 2 cycle reads and do 2-cycle inputs
//		M006	3-24-09	Added mask instruction
//		M007	3-26-09	make program memory external to allow data access (remove PGM instruction)
//		M007	4-28-09	fix iord strobe bug
//		M008	7-13-09	changed link flag to overflow flag
//		M009	7-18-09	16-bit register load now 2 sequential 8-bit load instructions
//
module CPU16(ioaddr, din, iord, dout, iowr,
				 maddr, min, mout, mwr,
				 paddr, pdata, pcs,
				 clk, rst);
	 output [6:0] ioaddr;		// I/O address bus (modifier field plus B addr. field)
    input [15:0] din;			// I/O data input
	 output iord;					// requesting data on input bus
    output [15:0] dout;			// I/O data output
    output iowr;					// valid data on output bus
	 output [13:0] maddr;		// memory address
	 input [15:0] min;			// memory data input
	 output [15:0] mout;			// memory data output
	 output mwr;					// memory write
	 output [13:0] paddr;		// program memory address
	 input [15:0] pdata;			// program data (instructions)
	 output pcs;					// program memory enable (used to hold instruction for additional cycle)
	 input clk;						// common clock
	 input rst;						// reset CPU state
// internal signals
wire [1:0] typ;					// instruction type
wire [2:0] op,mod;				// ALU operation and modifier
wire [2:0] fmux;					// ALU function multiplexer select input
wire [3:0] aa,ab;					// register address fields
wire [15:0] da,db,f;				// data busses
wire [3:0] we;						// write to register A (by nibble)
wire zero,carry,ovflo;			// ALU flags
wire aluwo;							// write-only operations
wire aluop;							// ALU operations
wire mvi,adi,cpi;					// immediate data instructions
wire inp,out,memop,crd,cwr;	// I/O and memory operations
wire inrd1;							// I/O input or memory read cycle 1 and 2
reg inrd2;
wire [15:0] idata;				// 16-bit immediate data
wire [7:0] id;						// 8-bit immediate data
reg pid;								// last instruction was immediate data
reg [3:0] preg;					// last A register accessed
// Program flow control logic reads instructions from ROM and controls branching.
// Branches occur if specific flags set - MSB flag input is unconditional branch.
PC14 pfcu (
	.flags({carry,zero,ovflo,1'b1}),	// status flags
	.paddr(paddr),							// address of next intruction
	.pdata(pdata),							// current instruction
	.rdata(da[13:0]),						// data from A register
	.ce(~inrd1),							// PC count enable (low for 1st cycle of 2-cycle instructions)
	.clk(clk),
	.rst(rst)
	);
assign pcs = ~inrd1;	// enable program RAM output
// assign fields for load and data manipulation instuctions (program control decoded elsewhere)
assign aa = pdata[3:0];								// A register address
assign ab = pdata[7:4];								// B register address
assign mod = pdata[11:8];							// operation modifier
assign cwr = pdata[8];								// write to memory
assign crd = pdata[9];								// read from memory
assign op = pdata[13:11];							// operation code field
assign typ = pdata[15:14];							// instruction type field
assign mvi = (pdata[15:12] == 4'b0111);		// Load immediate instruction
assign adi = (pdata[15:12] == 4'b1000);		// add imediate instruction
assign cpi = (pdata[15:12] == 4'b1001);		// compare immediate instruction
assign id = pdata[11:4];							// immediate data field
assign out = (typ == 2) & (op == 7);			// Output operation
assign inp = (typ == 3) & (op == 0);			// Input operation
assign memop = (typ == 3) & (op == 1);			// memory operation
assign aluop = (typ == 3);							// ALU read/write operation
assign aluwo = (typ == 2);							// write-only ALU operations
// decode input, memory read or memory swap instruction cycle 1 and cycle 2
assign inrd1 = aluop & (op[2:1] == 0) & (crd|~op[0]) & ~inrd2;
always @ (posedge clk) if (rst) inrd2 <= 0; else inrd2 <= inrd1;
// ALU does arithmetic, logical and shifting operations in 1 clock cycle
// first MVI loads LSB and extends sign and second MVI loads MSB without modifying LSB
always @ (posedge clk)
begin
	pid <= mvi;	// remember last loading of immediate data
	preg <= aa;	// remember last register loaded
end
assign idata = (pid & (preg == aa)) ? {id,da[7:0]} : {id[7],id[7],id[7],id[7],id[7],id[7],id[7],id[7],id};
// modify operation field to control ALU output multiplexer
// operation field values of 2 and 3 select port 2 for 16-bit register-to-register arithmetic
// ADI and CPI instructions select port 3 to get output of constant adder
// MVI instruction selects port 7 to load immediate data
assign fmux = (mvi|adi|cpi) ? {mvi,mvi|adi|cpi,mvi|adi|cpi} : ((op == 3'b011) ? 3'b010 : op);
ALU16 alu (
	.a(da),							// 1st operand from register
	.b(db),							// 2nd operand from register
	.c(id),							// 2nd operand from instruction field
	.d0(din),						// data from I/O port
	.d1(min),						// data from memory
	.d7(idata),						// immediate data
	.iv(aluop),						// ALU data valid
	.wov(aluwo),					// ALU write-only operations
	.op(fmux),						// select ALU operation
	.mod({(op != 3) & mod[2],mod[1:0]}),	// operation modifier - suppress carry propogation on 4 and 8-bit math
	.csub(pdata[12]),				// immediate operation modifier
	.f(f),							// ALU output
	.zero(zero),					// zero, carry and overflow flags from ALU
	.carry(carry),
	.ovflo(ovflo),
	.clk(clk),						// for carry flag
	.rst(rst)
	);
// Dual-port memory containing data (A/B) and address (B) registers
// update registers on all ALU operations except memory writes and
// on move and add immediate instructions (2nd cycle of memory reads)
assign we[0] = mvi | adi | (aluop & (op[2:1] != 0)) | inrd2;
assign we[1] = we[0] & ((op != 3) | ~mod[2]);	// nibble-wide operations excluded
assign we[2] = we[1] & (op != 3);	// byte and nibble-wide operations excluded
assign we[3] = we[2];
RAM16X16D regfile (
	.AA(aa),		// 1st operand address
	.AB(ab),		// 2nd operand address
	.DA(f),		// 1st operand input
	.YA(da),		// 1st operand output
	.YB(db),		// 2nd operand output
	.WE(we),		// write results to high aand low bytes
	.WCLK(clk)	// write clock (read is asynchronous)
	);
// Memory Bus
assign maddr = db[13:0];	// 14-bit memory address
assign mout = da;				// 16-bit data from register A
assign mwr = memop & cwr;	// write to memory
// I/O Bus
assign ioaddr = {mod,ab};	// 7-bit I/O address
assign dout = da;				// 16-bit data from register A
assign iord = inp & inrd1;	// read from input port (strobe on 1st cycle only)
assign iowr = out;			// write to output port
endmodule
